package be.iminds.ilabt.jfed.experimenter_gui.util.ui;

import javafx.animation.AnimationTimer;
import javafx.scene.control.TextArea;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingDeque;

/**
 * This class is used to process large amounts of messages received by a text-area.
 *
 * By using an animation timer, we prevent concurrency problems in the JavaFX-platform (see
 *
 * http://stackoverflow.com/questions/30863862/javafx-append-text-to-textarea-throws-exception and
 * http://stackoverflow.com/questions/31408363/javafx-changelistener-not-always-working/31414801#31414801
 *
 * )
 *
 * Created by Thijs on 5/5/2017.
 */
public class TextAreaMessageConsumer extends AnimationTimer {
    private final BlockingQueue<String> messageQueue = new LinkedBlockingDeque<>();
    private final TextArea textArea;

    public TextAreaMessageConsumer(TextArea textArea) {
        this.textArea = textArea;
    }

    public void addMessage(String message) {
        messageQueue.add(message);
    }

    @Override
    public void handle(long now) {
        StringBuilder newMessages = new StringBuilder();

        String message;
        while((message = messageQueue.poll()) != null){
            newMessages.append(message);
        }

        textArea.appendText(newMessages.toString());
    }
}
