package be.iminds.ilabt.jfed.rspec_fx;

import be.iminds.ilabt.jfed.rspec_fx.model.controller.ExperimentBarrierSegment;
import be.iminds.ilabt.jfed.rspec_fx.model.controller.ExperimentCommand;
import be.iminds.ilabt.jfed.rspec_fx.model.controller.Timestamp;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.request.ext.delay.Pipe;
import be.iminds.ilabt.jfed.rspec.request.ext.delay.SliverTypeShaping;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.stream.Location;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import static be.iminds.ilabt.jfed.rspec.parser.RspecXmlConstants.*;

public class StaxRspecFxParser {

    private static final Logger LOG = LoggerFactory.getLogger(StaxRspecFxParser.class);
    private static final StaxRspecFxParser INSTANCE = new StaxRspecFxParser();

    public static StaxRspecFxParser getInstance() {
        return INSTANCE;
    }

    private StaxRspecFxParser(){}

    public SliverTypeShaping parseSliverTypeShaping(Location location, XMLEventReader eventReader) throws RspecParseException {
        try {
            XMLEvent e = eventReader.nextEvent();

            assert e.isStartElement();
            StartElement startElement = e.asStartElement();
            assert startElement.getName().equals(Q_SLIVER_TYPE_SHAPING);

            SliverTypeShaping res = new SliverTypeShaping();

            while (eventReader.hasNext()) {
                e = eventReader.nextEvent();

                if (e.isStartElement()) {
                    if (e.asStartElement().getName().equals(Q_PIPE)) {
                        res.getPipe().add(parsePipe(e, eventReader));
                    }

                } else if (e.isEndElement()) {
                    if (e.asEndElement().getName().equals(Q_SLIVER_TYPE_SHAPING))
                        return res;
                } else {
                    LOG.warn("Ignoring event {} in SliverTypeShaping", e);
                    //throw new RspecParseException(location, "Unexpected content in ExperimentCommand");
                }
            }

        } catch (XMLStreamException e1) {
            throw new RspecParseException(location, "Could not parse SliverTypeShaping", e1);
        }


        throw new RspecParseException(location, "Did not encounter end element for SliverTypeShaping");

    }


    private Pipe parsePipe(XMLEvent e, XMLEventReader eventReader) throws RspecParseException {
        try {
            Pipe pipe = new Pipe();

            assert e.isStartElement();
            StartElement startElement = e.asStartElement();
            assert startElement.getName().equals(Q_PIPE);

            Attribute sourceAtt = startElement.getAttributeByName(Q_ATT_SOURCE);
            if (sourceAtt != null) {
                pipe.setSource(sourceAtt.getValue());
            }
            Attribute destAtt = startElement.getAttributeByName(Q_ATT_DEST);
            if (destAtt != null) {
                pipe.setDest(destAtt.getValue());
            }

            Attribute capacityAtt = startElement.getAttributeByName(Q_ATT_CAPACITY);
            if (capacityAtt != null) {
                pipe.setCapacity(Integer.parseInt(capacityAtt.getValue()));
            }
            Attribute latencyAtt = startElement.getAttributeByName(Q_ATT_LATENCY);
            if (latencyAtt != null) {
                pipe.setLatency(Integer.parseInt(latencyAtt.getValue()));
            }

            Attribute packetLossAtt = startElement.getAttributeByName(Q_ATT_PACKET_LOSS);
            if (packetLossAtt != null) {
                pipe.setPacketLoss(Float.parseFloat(packetLossAtt.getValue()));
            }

            Attribute gentleAtt = startElement.getAttributeByName(Q_ATT_GENTLE);
            if (gentleAtt != null) {
                pipe.setGentle(Boolean.parseBoolean(gentleAtt.getValue()));
            }
            Attribute redAtt = startElement.getAttributeByName(Q_ATT_RED);
            if (redAtt != null) {
                pipe.setRed(Boolean.parseBoolean(redAtt.getValue()));
            }
            Attribute queueInBytesAtt = startElement.getAttributeByName(Q_ATT_QUEUE_IN_BYTES);
            if (queueInBytesAtt != null) {
                pipe.setQueueInBytes(Boolean.parseBoolean(queueInBytesAtt.getValue()));
            }
            Attribute limitAtt = startElement.getAttributeByName(Q_ATT_LIMIT);
            if (limitAtt != null) {
                pipe.setLimit(Integer.parseInt(limitAtt.getValue()));
            }
            Attribute maxtreshAtt = startElement.getAttributeByName(Q_ATT_MAXTRESH);
            if (maxtreshAtt != null) {
                pipe.setMaxthresh(Integer.parseInt(maxtreshAtt.getValue()));
            }
            Attribute treshAtt = startElement.getAttributeByName(Q_ATT_TRESH);
            if (treshAtt != null) {
                pipe.setThresh(Integer.parseInt(treshAtt.getValue()));
            }
            Attribute lintermAtt = startElement.getAttributeByName(Q_ATT_LINTERM);
            if (lintermAtt != null) {
                pipe.setLinterm(Integer.parseInt(lintermAtt.getValue()));
            }

            Attribute qWeightAtt = startElement.getAttributeByName(Q_ATT_Q_WEIGHT);
            if (qWeightAtt != null) {
                pipe.setQWeight(Float.parseFloat(qWeightAtt.getValue()));
            }


            while (eventReader.hasNext()) {
                e = eventReader.nextEvent();

                if (e.isEndElement()) {
                    if (e.asEndElement().getName().equals(Q_PIPE))
                        return pipe;
                } else {
                    LOG.warn("Ignoring event {} in pipe", e);
                    //throw new RspecParseException(location, "Unexpected content in ExperimentCommand");
                }
            }

        } catch (XMLStreamException e1) {
            throw new RspecParseException(e.getLocation(), "Could not parse Pipe", e1);
        }


        throw new RspecParseException(e.getLocation(), "Did not encounter end element for Pipe");

    }

    public ExperimentBarrierSegment parseExperimentBarrierSegment(Location location, XMLEventReader eventReader) throws RspecParseException {

        ExperimentBarrierSegment res = new ExperimentBarrierSegment();
        try {
            XMLEvent e = eventReader.nextEvent();

            assert e.isStartElement();
            StartElement startElement = e.asStartElement();
            assert startElement.getName().equals(Q_JFEDCOMMAND_EXP_BARRIER_SEG);

            int orderNumber;
            Attribute orderNumberAtt = startElement.getAttributeByName(Q_ATT_ORDER_NUMBER);
            if (orderNumberAtt != null) {
                orderNumber = Integer.parseInt(orderNumberAtt.getValue());
            }else{
                throw new RspecParseException(startElement.getLocation(), "OrderNumber is a required attribute");
            }
            String tag = null;
            Attribute tagAtt = startElement.getAttributeByName(Q_ATT_TAG);
            if (tagAtt != null) {
                tag = tagAtt.getValue();
            }


            if (eventReader.hasNext()) {
                e = eventReader.nextEvent();

                if (e.isEndElement()) {
                    return new ExperimentBarrierSegment(tag, orderNumber);
                } else {
                    throw new RspecParseException(location, "Unexpected content in ExperimentBarrierSegment");
                }
            }

        } catch (XMLStreamException e1) {
            throw new RspecParseException(location, "Could not parse ExperimentBarrierSegment", e1);
        }


        return res;

    }

    public ExperimentCommand parseExperimentCommand(Location location, XMLEventReader eventReader) throws RspecParseException {
        try {
            XMLEvent e = eventReader.nextEvent();

            assert e.isStartElement();
            StartElement startElement = e.asStartElement();
            assert startElement.getName().equals(Q_JFEDCOMMAND_EXP_COMMAND) || startElement.getName().equals(Q_JFEDCOMMAND_EXP_COMMAND_OLDERROR);

            int orderNumber;
            Attribute orderNumberAtt = startElement.getAttributeByName(Q_ATT_BARRIER_SEGMENT_ORDER_NUMBER);
            if (orderNumberAtt != null) {
                orderNumber = Integer.parseInt(orderNumberAtt.getValue());
            }else{
                throw new RspecParseException(startElement.getLocation(), "OrderNumber is a required attribute");

            }
            String tag = null;
            Attribute tagAtt = startElement.getAttributeByName(Q_ATT_TAG);
            if (tagAtt != null) {
                tag = tagAtt.getValue();
            }


            String command = null;
            Timestamp duration = null;
            Timestamp startingTime = null;
            while (eventReader.hasNext()) {
                e = eventReader.nextEvent();

                if (e.isStartElement()) {
                    if (e.asStartElement().getName().equals(Q_JFEDCOMMAND_COMMAND)) {
                        command = eventReader.getElementText();
                    }
                    if (e.asStartElement().getName().equals(Q_JFEDCOMMAND_DURATION)) {
                        duration = parseCommandTime(e, eventReader);
                    }
                    if (e.asStartElement().getName().equals(Q_JFEDCOMMAND_STARTING_TIME)) {
                        startingTime = parseCommandTime(e, eventReader);
                    }
                } else if (e.isEndElement()) {
                    if (e.asEndElement().getName().equals(Q_JFEDCOMMAND_EXP_COMMAND)
                            || e.asEndElement().getName().equals(Q_JFEDCOMMAND_EXP_COMMAND_OLDERROR))
                        return new ExperimentCommand(tag, null, orderNumber, command, startingTime, duration);
                } else {

                    //throw new RspecParseException(location, "Unexpected content in ExperimentCommand");
                }
            }

        } catch (XMLStreamException e1) {
            throw new RspecParseException(location, "Could not parse ExperimentCommand", e1);
        }


        throw new RspecParseException(location, "Did not encounter end element for ExperimentCommand");

    }


    private Timestamp parseCommandTime(XMLEvent e, XMLEventReader eventReader) throws XMLStreamException, RspecParseException {
        while(eventReader.hasNext()){
            e = eventReader.nextEvent();

            if(e.isStartElement() && e.asStartElement().getName().equals(Q_JFEDCOMMAND_TIME)) {
                try {
                    return new Timestamp(Double.parseDouble(eventReader.getElementText()));
                } catch (NumberFormatException ex) {
                    throw new RspecParseException("Expected a number as time", ex);
                }
            }else {
                if(e.isEndElement()){
                    throw new RspecParseException(e.getLocation(), "Could not find timestamp to parse");
                }
            }
        }

        throw new RspecParseException(e.getLocation(), "Could not find timestamp to parse");
    }
}
